import json
from typing import Any, Dict, List, Union, Tuple, Optional
from confio.core.base import ConfItem, IConf, logger


class ConfDameng(IConf):
    """
    基于达梦数据库的配置存储
    """
    SQL = {
        # 查询表是否存在
        'table_exists': """SELECT COUNT(*) as "count" FROM ALL_TABLES WHERE OWNER = '{db_name}' AND TABLE_NAME = '{table_name}'""",
        # 创建表结构
        'create_table': """CREATE TABLE "{db_name}"."{table_name}" (
        "id" VARCHAR(128) PRIMARY KEY NOT NULL,
        "name" VARCHAR(64),
        "type" VARCHAR(32) DEFAULT '',
        "value" TEXT,
        "user_value" TEXT,
        "value_type" VARCHAR(32),
        "enabled" INT DEFAULT 1,
        "desc" VARCHAR(256)
        )""",
        # 加载所有数据
        'load_all': 'SELECT * FROM "{table_name}"',
        # 通过id查询一项数据
        'get_by_id': 'SELECT * FROM "{table_name}" WHERE "id"=?',
        # 加载匹配指定前缀的数据
        'match_items': 'SELECT * FROM "{table_name}" WHERE "id" LIKE ?',
        # 更新数据值
        'update_sys_value_by_id': 'UPDATE "{table_name}" SET "value"=?, "value_type"=? WHERE "id"=?',
        'update_user_value_by_id': 'UPDATE "{table_name}" SET "user_value"=?, "value_type"=? WHERE "id"=?',
        # 更新所有字段
        'update_by_id': 'UPDATE "{table_name}" SET '
                        '"name"=?,'
                        '"value"=?,'
                        '"user_value"=?,'
                        '"value_type"=?,'
                        '"type"=?,'
                        '"desc"=?,'
                        '"enabled"=?'
                        ' WHERE "id"=?',
        # 添加项
        'add_item': 'INSERT INTO "{table_name}"'
                    '("id","name","value","user_value","value_type","type","desc","enabled")'
                    'VALUES(?,?,?,?,?,?,?,?)',
        # 移除项
        'remove_item': 'DELETE FROM "{table_name}" WHERE "id"=?',
        # 清空数据
        'clear_table': 'TRUNCATE TABLE "{table_name}"'
    }

    def __init__(self,
                 schema: str,
                 host='127.0.0.1',
                 port=5236,
                 user='root',
                 password='',
                 table_name='conf_items',
                 **pool_args):
        """
        :pool_args: 指定连接池的一些参数，如 mincached maxcached maxshared maxconnections blocking setsession ping 等
        """

        self.table_name = table_name
        self.con_option = {
            'schema': schema,
            'host': host,
            'port': port,
            'user': user,
            'password': password,
        }
        # logger.debug(
        #     'Initialize database configuration with table "{table_name}" via URI '
        #     'dm://{host}:{port}/{database}'.format(
        #         table_name=table_name,
        #         **self.con_option
        #     )
        # )
        import dmPython
        from dbutils.pooled_db import PooledDB
        pool_options = {
            'creator': dmPython,
            'mincached': 5,
            'maxcached': 20,
            'maxshared': 10,
            'maxconnections': 20,
            'blocking': True,
            'maxusage': 1000,
            'setsession': [],
            'ping': 0,
            **self.con_option,
            **pool_args
        }
        self.pool = PooledDB(**pool_options)
        self._tested = False
        super(ConfDameng, self).__init__()

    def _execute(self, sql: str, *args):
        if not self._tested:
            self._tested = True
            if not self._table_exists():
                self._create_table()

        conn = self.pool.connection()
        cursor = None
        is_query = sql.startswith('SELECT')
        try:
            cursor = conn.cursor()
            cursor.execute(sql, args)
            if is_query:
                cols = cursor.column_names
                data = []
                for row in cursor.fetchall():
                    item = {}
                    i = 0
                    for col in cols:
                        item[col] = row[i]
                        i += 1
                    data.append(item)

            else:
                data = None
        except Exception as e:
            print(str(e))
            return 0, None
        else:
            if not is_query:
                conn.commit()
        finally:
            if cursor is not None:
                cursor.close()
            conn.close()

        return cursor.rowcount, data

    def _table_exists(self) -> bool:
        _, data = self._execute(self.SQL['table_exists'].format(
            db_name=self.con_option['schema'],
            table_name=self.table_name
        ))
        return data[0]['count'] > 0

    def _create_table(self):
        create_table_sql = self.SQL['create_table'].format(
            table_name=self.table_name
        )
        create_table_sql = create_table_sql.split('ENGINE=InnoDB DEFAULT CHARSET=utf8')[0].rstrip()
        self._execute(create_table_sql)

    def load(self) -> List[ConfItem]:
        rows, data = self._execute(self.SQL['load_all'].format(
            table_name=self.table_name
        ))
        return [
            self.row2item(data_item)
            for data_item in data
        ]

    def get(self, id_: str, default=None, value_only=True) -> Union[ConfItem, Any]:
        rows, data = self._execute(self.SQL['get_by_id'].format(
            table_name=self.table_name
        ), id_)

        if rows:
            item = self.row2item(data[0])
            return item.value if value_only else item
        if id_ != self.VERSION_KEY:
            logger.warning('Conf item "%s" not found' % id_, stack_info=self.STACK_INFO)
        return default

    def match(self, prefix: str, value_only=True, fullkey=False) -> Dict[str, Union[ConfItem, Any]]:
        if prefix[-1] != '.':
            prefix = prefix + '.'
        prefix_len = len(prefix)

        rows, data = self._execute(self.SQL['match_items'].format(
            table_name=self.table_name
        ), prefix + '%')

        items = {}
        for data_item in data:
            item = self.row2item(data_item)
            items[item.id if fullkey else item.id[prefix_len:]] = item.value if value_only else item

        return items

    def set(self, id_: str, value, update_sys_value=False, allow_add=False):
        rows, _ = self._execute(self.SQL['get_by_id'].format(
            table_name=self.table_name
        ), id_)
        if not rows:
            if not allow_add:
                logger.warning('Conf item "%s" not found' % id_, stack_info=self.STACK_INFO)
                return
            self._add_item(id_, value, update_sys_value)
            return

        self._set_item(id_, value, update_sys_value)

        if id_ != self.VERSION_KEY:
            self.version(True)

    def batch_set(self, items: Union[Dict[str, Any], List[ConfItem]], prefix: str = None, update_sys_value=False,
                  allow_add=False):
        if prefix:
            if prefix[-1] != '.':
                prefix = prefix + '.'
        else:
            prefix = ''

        if isinstance(items, list):
            items = {
                item.id: item
                for item in items
            }

        all_items = self.load()
        data_map = {}

        for item in all_items:
            data_map[item.id] = item

        for key in items:
            id_ = prefix + key
            if id_ == self.VERSION_KEY:
                continue
            item = items[key]
            if id_ in data_map:
                self._set_item(id_, item, update_sys_value)
                continue
            if not allow_add:
                logger.warning('Conf item "%s" not found' % id_, stack_info=self.STACK_INFO)
                continue
            self._add_item(id_, item, update_sys_value)

        self.version(True)

    def remove(self, id_: Union[str, List[str], Tuple[str]]) -> Optional[Union[ConfItem, List[ConfItem]]]:
        is_list = isinstance(id_, (list, tuple))
        if is_list:
            id_list = id_
        else:
            id_list = (id_,)

        all_items = self.load()
        data_map = {}

        for item in all_items:
            data_map[item.id] = item

        remove_list = []

        for id_ in id_list:
            if id_ not in data_map:
                logger.warning('Conf item "%s" not found' % id_, stack_info=self.STACK_INFO)
                continue
            self._execute(self.SQL['remove_item'].format(
                table_name=self.table_name
            ), id_)
            remove_list.append(data_map[id_])

        self.version(True)

        if is_list:
            return remove_list
        return None if len(remove_list) == 0 else remove_list[0]

    def clear(self):
        """
        清空配置表
        :return:
        """
        self._execute(self.SQL['clear_table'].format(
            table_name=self.table_name
        ))

    def row2item(self, model: dict) -> ConfItem:
        value_type = model['value_type']
        id_ = model['id']

        item = ConfItem()
        item.id = id_
        item.name = model['name']
        item.raw_sys_value = self._load_value(id_, value_type, model['value'])
        item.raw_user_value = self._load_value(id_, value_type, model['user_value'])
        item.type = item.parser.cast_enum(model['type'])
        item.desc = model['desc']
        item.enabled = model['enabled'] == 1

        return item

    def _set_item(self, id_, value, update_sys_value: bool):
        if isinstance(value, ConfItem):
            sys_value = value.raw_sys_value
            user_value = value.raw_user_value

            value_type = self._check_value_type(sys_value, user_value, id_)

            sys_value = self._dump_value(sys_value)
            user_value = self._dump_value(user_value)

            self._execute(self.SQL['update_by_id'].format(
                table_name=self.table_name
            ), value.name, sys_value, user_value, value_type,
                value.type.value, value.desc, value.enabled, id_)
            return

        dumped_value = self._dump_value(value)
        if update_sys_value:
            value_type = self._get_value_type(value)
            sql = self.SQL['update_sys_value_by_id'].format(
                table_name=self.table_name
            )
            self._execute(sql, dumped_value, value_type, id_)
        else:
            value_type = self._check_value_type(self.get(id_, value_only=False).raw_sys_value, value, id_)
            sql = self.SQL['update_user_value_by_id'].format(
                table_name=self.table_name
            )
            self._execute(sql, dumped_value, value_type, id_)

    def _add_item(self, id_, value, update_sys_value):
        if isinstance(value, ConfItem):
            sys_value = value.raw_sys_value
            user_value = value.raw_user_value
            item = value
            item.value_type = self._check_value_type(sys_value, user_value, id_)
            item.raw_sys_value = self._dump_value(sys_value)
            item.raw_user_value = self._dump_value(user_value)
        else:
            item = ConfItem()
            item.value_type = self._get_value_type(value)
            item.update(self._dump_value(value), update_sys_value)

        item.id = id_

        self._execute(self.SQL['add_item'].format(
            table_name=self.table_name
        ), item.id, item.name, item.raw_sys_value, item.raw_user_value, item.value_type,
            item.type.value, item.desc, item.enabled)

    @classmethod
    def _get_value_type(cls, *values):
        for val in values:
            if val is not None:
                return type(val).__name__

        return None

    @classmethod
    def _check_value_type(cls, sys_val, user_val, id_):
        sys_type = cls._get_value_type(sys_val)
        user_type = cls._get_value_type(user_val)

        if sys_type is None:
            # 系统值类型为None，则不检查了
            return user_type

        if user_type is None:
            # 用户值类型为None，则不检查了
            return sys_type

        if sys_type == user_type:
            return sys_type

        if sys_type == 'bool':
            if user_val in ('true', 'false'):
                return sys_type

        if sys_type == 'int':
            try:
                int(user_val)
                return sys_type
            except:
                pass

        if sys_type == 'float':
            try:
                float(user_val)
                return sys_type
            except:
                pass

        if sys_type in ('list', 'dict', 'tuple'):
            try:
                json.loads(user_val)
                return sys_type
            except:
                pass

        raise Exception('Invalid value type %r for item %r, type %r required.' % (user_type, id_, sys_type))

    @classmethod
    def _load_value(cls, id_, value_type, value):
        if value_type is None or value_type == 'NoneType':
            return value

        if value is None or value == '':
            return value

        if value_type == 'int':
            return int(value)

        if value_type == 'bool':
            return value.lower() == 'true'

        if value_type == 'float':
            return float(value)

        if value_type in (dict, list):
            return json.loads(value)

        if value_type == 'str':
            return value

        if value_type in ('list', 'dict', 'tuple'):
            if value is None:
                return value

            return json.loads(value)

        logger.error('Cannot resolve value type "%s" of "%s"' % (value_type, id_))
        return value

    @classmethod
    def _dump_value(cls, value):
        if isinstance(value, bool):
            value = 'true' if value else 'false'
        elif isinstance(value, (dict, list, tuple)):
            value = None if value is None else json.dumps(value, ensure_ascii=False)

        return value
